from ..unit.convertor import Convertor


# the value indicating that temperature value is expressed in some unit
TemperatureFahrenheit = 50
TemperatureCelsius = 51
TemperatureKelvin = 52
TemperatureRankin = 53


class Temperature(Convertor):
    """ Temperature object keeps temperature or speed values """

    __name__ = 'Temperature'

    _units = {
        TemperatureFahrenheit: {'name': '°F', 'accuracy': 1,
                                'to': lambda v: v,
                                'from': lambda v: v},
        TemperatureRankin: {'name': '°R', 'accuracy': 1,
                            'to': lambda v: v - 459.67,
                            'from': lambda v: v + 459.67},
        TemperatureCelsius: {'name': '°C', 'accuracy': 1,
                             'to': lambda v: v * 9 / 5 + 32,
                             'from': lambda v: (v - 32) * 5 / 9},
        TemperatureKelvin: {'name': '°K', 'accuracy': 1,
                            'to': lambda v: (v - 273.15) * 9 / 5 + 32,
                            'from': lambda v: (v - 32) * 5 / 9 + 273.15},
    }



import math
from .bmath import unit, Vector
from .projectile import Ammunition
from .weapon import Weapon, TwistLeft
from .atmosphere import Atmosphere
from .shot_parameters import ShotParameters
from .wind import WindInfo
from .trajectory_data import TrajectoryData, Timespan

cZeroFindingAccuracy = 0.000005
cMinimumVelocity = 50.0
cMaximumDrop = -15000
cMaxIterations = 10
cGravityConstant = -32.17405


class TrajectoryCalculator(object):
    """
    TrajectoryCalculator table is used to calculate the trajectory
    of a projectile shot with the parameters specified
    """

    def __init__(self):
        self._maximum_calculator_step_size = unit.Distance(1, unit.DistanceFoot).validate()

    @property
    def maximum_calculator_step_size(self) -> unit.Distance:
        """
        :return: the maximum size of one calculation iteration.
        """
        return self._maximum_calculator_step_size

    @maximum_calculator_step_size.setter
    def maximum_calculator_step_size(self, value: unit.Distance):
        """
        Sets the maximum size of one calculation iteration.
        As the generic rule, the maximum step of the calculation must not be greater than
        a half of the step used in the short parameter. The smaller value is, the calculation is more precise but
        takes more time to calculate. From practical standpoint the value in range from 0.5 to 5 feet produces
        good enough accuracy.
        :param value: unit.Distance instance
        :return: None
        """
        self._maximum_calculator_step_size = value

    def get_calculation_step(self, step: float) -> float:
        step = step / 2  # do it twice for increased accuracy of velocity calculation and 10 times per step
        maximum_step: float = self._maximum_calculator_step_size.get_in(unit.DistanceFoot)

        if step > maximum_step:
            step_order = int(math.floor(math.log10(step)))
            maximum_order = int(math.floor(math.log10(maximum_step)))
            step = step / math.pow(10, float(step_order - maximum_order + 1))

        return step

    def sight_angle(self, ammunition: Ammunition, weapon: Weapon, atmosphere: Atmosphere) -> unit.Angular:
        """
        The calculated value is to be used as sightAngle parameter of the ShotParameters structure
        :param ammunition: Ammunition instance
        :param weapon: Weapon instance
        :param atmosphere: Atmosphere instance
        :return: sight angle for a rifle with scope height specified and zeroed using the ammo specified at
        the range specified and under the conditions (atmosphere) specified
        """

        calculation_step = self.get_calculation_step(
            unit.Distance(10, weapon.zero.zero_distance.units).validate().get_in(unit.DistanceFoot))

        mach = atmosphere.mach.get_in(unit.VelocityFPS)
        density_factor = atmosphere.density_factor
        muzzle_velocity = ammunition.muzzle_velocity.get_in(unit.VelocityFPS)
        barrel_azimuth = 0.0
        barrel_elevation = 0

        iterations_count = 0

        zero_finding_error = cZeroFindingAccuracy * 2
        gravity_vector = Vector(0, cGravityConstant, 0)
        while zero_finding_error > cZeroFindingAccuracy and iterations_count < cMaxIterations:
            velocity = muzzle_velocity
            time = 0.0

            # x - distance towards target,
            # y - drop and
            # z - windage

            range_vector = Vector(0.0, -weapon.sight_height.get_in(unit.DistanceFoot), 0)
            velocity_vector = Vector(
                math.cos(barrel_elevation) * math.cos(barrel_azimuth),
                math.sin(barrel_elevation),
                math.cos(barrel_elevation) * math.sin(barrel_azimuth)
            ).multiply_by_const(velocity)

            zero_distance: float = weapon.zero.zero_distance.get_in(unit.DistanceFoot)
            maximum_range = zero_distance + calculation_step

            while range_vector.x <= maximum_range:
                if velocity < cMinimumVelocity or range_vector.y < cMaximumDrop:
                    break

                delta_time = calculation_step / velocity_vector.x
                velocity = velocity_vector.magnitude()
                drag = density_factor * velocity * ammunition.bullet.ballistic_coefficient.drag(velocity / mach)
                velocity_vector = velocity_vector.subtract(
                    (
                        velocity_vector.multiply_by_const(drag).subtract(gravity_vector)
                    ).multiply_by_const(delta_time)
                )
                delta_range_vector = Vector(calculation_step, velocity_vector.y * delta_time,
                                            velocity_vector.z * delta_time)
                range_vector = range_vector.add(delta_range_vector)
                velocity = velocity_vector.magnitude()
                time = time + delta_range_vector.magnitude() / velocity
                if math.fabs(range_vector.x - zero_distance) < 0.5 * calculation_step:
                    zero_finding_error = math.fabs(range_vector.y)
                    barrel_elevation = barrel_elevation - range_vector.y / range_vector.x
                    break

                iterations_count += 1
        return unit.Angular(barrel_elevation, unit.AngularRadian).validate()

    def trajectory(self, ammunition: Ammunition, weapon: Weapon, atmosphere: Atmosphere,
                   shot_info: ShotParameters, wind_info: list[WindInfo]) -> list[TrajectoryData]:
        """
        Calculates the trajectory with the parameters specified
        :param ammunition: Ammunition instance
        :param weapon: Weapon instance
        :param atmosphere: Atmosphere instance
        :param shot_info: ShotParameters instance
        :param wind_info: list[WindInfo]
        :return: trajectory with the parameters specified
        """
        range_to: float = shot_info.maximum_distance.get_in(unit.DistanceFoot)
        step: float = shot_info.step.get_in(unit.DistanceFoot)

        calculation_step = self.get_calculation_step(step)

        bullet_weight = ammunition.bullet.bullet_weight.get_in(unit.WeightGrain)

        stability_coefficient = 1.0

        calculate_drift: bool = False

        if weapon.has_twist and ammunition.bullet.has_dimensions:
            stability_coefficient = self.calculate_stability_coefficient(
                ammunition, weapon, atmosphere
            )
            calculate_drift = True

        ranges_length = int(math.floor(range_to / step)) + 1
        ranges = []

        barrel_azimuth = .0
        barrel_elevation = shot_info.sight_angle.get_in(unit.AngularRadian)
        barrel_elevation = barrel_elevation + shot_info.shot_angle.get_in(unit.AngularRadian)
        alt0: float = atmosphere.altitude.get_in(unit.DistanceFoot)

        # Never used in upstream, uncomment on need
        # density_factor, mach = atmosphere.get_density_factor_and_mach_for_altitude(alt0)

        current_wind = 0
        next_wind_range = 1e7

        if len(wind_info) < 1:
            wind_vector = Vector(0, 0, 0)
        else:
            if len(wind_info) > 1:
                next_wind_range = wind_info[0].until_distance.get_in(unit.DistanceFoot)
            wind_vector = self.wind_to_vector(shot_info, wind_info[0])

        muzzle_velocity = ammunition.muzzle_velocity.get_in(unit.VelocityFPS)
        gravity_vector = Vector(0, cGravityConstant, 0)
        velocity = muzzle_velocity
        time = .0

        # x - distance towards target,
        # y - drop and
        # z - windage

        range_vector = Vector(.0, -weapon.sight_height.get_in(unit.DistanceFoot), 0)
        velocity_vector = Vector(math.cos(barrel_elevation) * math.cos(barrel_azimuth), math.sin(barrel_elevation),
                                 math.cos(barrel_elevation) * math.sin(barrel_azimuth)).multiply_by_const(velocity)

        current_item = 0

        maximum_range = range_to
        next_range_distance = 0

        twist_coefficient = .0

        if calculate_drift:
            if weapon.twist.direction == TwistLeft:
                twist_coefficient = 1
            else:
                twist_coefficient = -1

        while range_vector.x <= maximum_range + calculation_step:
            if velocity < cMinimumVelocity or range_vector.y < cMaximumDrop:
                break

            density_factor, mach = atmosphere.get_density_factor_and_mach_for_altitude(alt0 + range_vector.y)

            if range_vector.x >= next_wind_range:
                current_wind += 1
                wind_vector = self.wind_to_vector(shot_info, wind_info[current_wind])

                if current_wind == len(wind_info) - 1:
                    next_wind_range = 1e7
                else:
                    next_wind_range = wind_info[current_wind].until_distance.get_in(unit.DistanceFoot)

            if range_vector.x >= next_range_distance:
                windage: float = range_vector.z
                if calculate_drift:
                    windage += (1.25 * (stability_coefficient + 1.2) * math.pow(time, 1.83) * twist_coefficient) / 12

                drop_adjustment = self.get_correction(range_vector.x, range_vector.y)
                windage_adjustment = self.get_correction(range_vector.x, windage)

                ranges.append(TrajectoryData(
                    time=Timespan(time),
                    travel_distance=unit.Distance(range_vector.x, unit.DistanceFoot).validate(),
                    drop=unit.Distance(range_vector.y, unit.DistanceFoot).validate(),
                    drop_adjustment=unit.Angular(drop_adjustment, unit.AngularRadian).validate(),
                    windage=unit.Distance(windage, unit.DistanceFoot).validate(),
                    windage_adjustment=unit.Angular(windage_adjustment, unit.AngularRadian),
                    velocity=unit.Velocity(velocity, unit.VelocityFPS),
                    mach=velocity / mach,
                    energy=unit.Energy(self.calculate_energy(bullet_weight, velocity),
                                       unit.EnergyFootPound),
                    optimal_game_weight=unit.Weight(self.calculate_ogv(bullet_weight, velocity),
                                                    unit.WeightPound))
                              )

                next_range_distance += step
                current_item += 1
                if current_item == ranges_length:
                    break

            delta_time = calculation_step / velocity_vector.x
            velocity_adjusted = velocity_vector.subtract(wind_vector)
            velocity = velocity_adjusted.magnitude()

            drag = density_factor * velocity * ammunition.bullet.ballistic_coefficient.drag(velocity / mach)
            velocity_vector = velocity_vector.subtract(
                (velocity_adjusted.multiply_by_const(drag).subtract(gravity_vector)).multiply_by_const(delta_time)
            )
            delta_range_vector = Vector(
                calculation_step, velocity_vector.y * delta_time, velocity_vector.z * delta_time
            )
            range_vector = range_vector.add(delta_range_vector)
            velocity = velocity_vector.magnitude()
            time = time + delta_range_vector.magnitude() / velocity

        return ranges

    @staticmethod
    def calculate_stability_coefficient(ammunition_info: Ammunition,
                                        rifle_info: Weapon, atmosphere: Atmosphere) -> float:
        weight: float = ammunition_info.bullet.bullet_weight.get_in(unit.WeightGrain)
        diameter: float = ammunition_info.bullet.bullet_diameter.get_in(unit.DistanceInch)
        twist: float = rifle_info.twist.twist.get_in(unit.DistanceInch) / diameter
        length: float = ammunition_info.bullet.bullet_length.get_in(unit.DistanceInch) / diameter
        sd = 30 * weight / (math.pow(twist, 2) * math.pow(diameter, 3) * length * (1 + math.pow(length, 2)))
        fv = math.pow(ammunition_info.muzzle_velocity.get_in(unit.VelocityFPS) / 2800, 1.0 / 3.0)
        ft: float = atmosphere.temperature.get_in(unit.TemperatureFahrenheit)
        pt: float = atmosphere.pressure.get_in(unit.PressureInHg)
        ftp = ((ft + 460) / (59 + 460)) * (29.92 / pt)

        return sd * fv * ftp

    @staticmethod
    def wind_to_vector(shot: ShotParameters, wind: WindInfo) -> Vector:
        sight_cosine = math.cos(shot.sight_angle.get_in(unit.AngularRadian))
        sight_sine = math.sin(shot.sight_angle.get_in(unit.AngularRadian))
        cant_cosine = math.cos(shot.cant_angle.get_in(unit.AngularRadian))
        cant_sine = math.sin(shot.cant_angle.get_in(unit.AngularRadian))
        range_velocity = wind.velocity.get_in(unit.VelocityFPS) * math.cos(wind.direction.get_in(unit.AngularRadian))
        cross_component = wind.velocity.get_in(unit.VelocityFPS) * math.sin(wind.direction.get_in(unit.AngularRadian))
        range_factor = -range_velocity * sight_sine
        return Vector(range_velocity * sight_cosine,
                      range_factor * cant_cosine + cross_component * cant_sine,
                      cross_component * cant_cosine - range_factor * cant_sine)

    @staticmethod
    def get_correction(distance: float, offset: float) -> [float, None]:
        if distance != 0:
            return math.atan(offset / distance)
        return

    @staticmethod
    def calculate_energy(bullet_weight: float, velocity: float) -> float:
        return bullet_weight * math.pow(velocity, 2) / 450400

    @staticmethod
    def calculate_ogv(bullet_weight: float, velocity: float) -> float:
        return math.pow(bullet_weight, 2) * math.pow(velocity, 3) * 1.5e-12



import unittest

from py_ballisticcalc.projectile import *
from py_ballisticcalc.drag import *
from py_ballisticcalc.weapon import *
from py_ballisticcalc.trajectory_calculator import *
from py_ballisticcalc.atmosphere import *
from py_ballisticcalc.trajectory_data import *
from py_ballisticcalc.shot_parameters import *
from py_ballisticcalc.bmath import unit


class TestPyBallisticCalc(unittest.TestCase):

    def test_zero1(self):
        bc = BallisticCoefficient(0.365, DragTableG1)
        projectile = Projectile(bc, unit.Weight(69, unit.WeightGrain).validate())
        ammo = Ammunition(projectile, unit.Velocity(2600, unit.VelocityFPS).validate())
        zero = ZeroInfo(unit.Distance(100, unit.DistanceYard).validate())
        weapon = Weapon(unit.Distance(3.2, unit.DistanceInch).validate(), zero)
        atmosphere = Atmosphere()
        calc = TrajectoryCalculator()

        sight_angle = calc.sight_angle(ammo, weapon, atmosphere)

        self.assertLess(math.fabs(sight_angle.get_in(unit.AngularRadian) - 0.001651), 1e-6,
                        f'TestZero1 failed {sight_angle.get_in(unit.AngularRadian):.10f}')

    def test_zero2(self):
        bc = BallisticCoefficient(0.223, DragTableG7)
        projectile = Projectile(bc, unit.Weight(168, unit.WeightGrain).validate())
        ammo = Ammunition(projectile, unit.Velocity(2750, unit.VelocityFPS).validate())
        zero = ZeroInfo(unit.Distance(100, unit.DistanceYard).validate())
        weapon = Weapon(unit.Distance(2, unit.DistanceInch).validate(), zero)
        atmosphere = Atmosphere()
        calc = TrajectoryCalculator()

        sight_angle = calc.sight_angle(ammo, weapon, atmosphere)

        self.assertLess(math.fabs(sight_angle.get_in(unit.AngularRadian) - 0.001228), 1e-6,
                        f'TestZero2 failed {sight_angle.get_in(unit.AngularRadian):.10f}')

    def assertEqualCustom(self, a, b, accuracy, name):
        with self.subTest():
            self.assertFalse(math.fabs(a - b) > accuracy, f'Assertion {name} failed ({a}/{b})')

    def validate_one(self, data: TrajectoryData, distance: float, velocity: float, mach: float, energy: float,
                     path: float, hold: float, windage: float, wind_adjustment: float, time: float, ogv: float,
                     adjustment_unit: int):

        self.assertEqualCustom(distance, data.travelled_distance.get_in(unit.DistanceYard), 0.001, "Distance")
        self.assertEqualCustom(velocity, data.velocity.get_in(unit.VelocityFPS), 5, "Velocity")
        self.assertEqualCustom(mach, data.mach_velocity, 0.005, "Mach")
        self.assertEqualCustom(energy, data.energy.get_in(unit.EnergyFootPound), 5, "Energy")
        self.assertEqualCustom(time, data.time.total_seconds, 0.06, "Time")
        self.assertEqualCustom(ogv, data.optimal_game_weight.get_in(unit.WeightPound), 1, "OGV")

        if distance >= 800:
            self.assertEqualCustom(path, data.drop.get_in(unit.DistanceInch), 4, 'Drop')
        elif distance >= 500:
            self.assertEqualCustom(path, data.drop.get_in(unit.DistanceInch), 1, 'Drop')
        else:
            self.assertEqualCustom(path, data.drop.get_in(unit.DistanceInch), 0.5, 'Drop')

        if distance > 1:
            self.assertEqualCustom(hold, data.drop_adjustment.get_in(adjustment_unit), 0.5, 'Hold')

        if distance >= 800:
            self.assertEqualCustom(windage, data.windage.get_in(unit.DistanceInch), 1.5, "Windage")
        elif distance >= 500:
            self.assertEqualCustom(windage, data.windage.get_in(unit.DistanceInch), 1, "Windage")
        else:
            self.assertEqualCustom(windage, data.windage.get_in(unit.DistanceInch), 0.5, "Windage")

        if distance > 1:
            self.assertEqualCustom(wind_adjustment, data.windage_adjustment.get_in(adjustment_unit), 0.5, "WAdj")

    def test_path_g1(self):
        bc = BallisticCoefficient(0.223, DragTableG1)
        projectile = Projectile(bc, unit.Weight(168, unit.WeightGrain).validate())
        ammo = Ammunition(projectile, unit.Velocity(2750, unit.VelocityFPS).validate())
        zero = ZeroInfo(unit.Distance(100, unit.DistanceYard).validate())
        weapon = Weapon(unit.Distance(2, unit.DistanceInch).validate(), zero)
        atmosphere = Atmosphere()
        shot_info = ShotParameters(unit.Angular(0.001228, unit.AngularRadian).validate(),
                                   unit.Distance(1000, unit.DistanceYard).validate(),
                                   unit.Distance(100, unit.DistanceYard).validate())
        wind = WindInfo.create_only_wind_info(unit.Velocity(5, unit.VelocityMPH).validate(),
                                              unit.Angular(-45, unit.AngularDegree).validate())

        calc = TrajectoryCalculator()
        data = calc.trajectory(ammo, weapon, atmosphere, shot_info, wind)

        self.assertEqualCustom(len(data), 11, 0.1, "Length")

        test_data = [
            [data[0], 0, 2750, 2.463, 2820.6, -2, 0, 0, 0, 0, 880, unit.AngularMOA],
            [data[1], 100, 2351.2, 2.106, 2061, 0, 0, -0.6, -0.6, 0.118, 550, unit.AngularMOA],
            [data[5], 500, 1169.1, 1.047, 509.8, -87.9, -16.8, -19.5, -3.7, 0.857, 67, unit.AngularMOA],
            [data[10], 1000, 776.4, 0.695, 224.9, -823.9, -78.7, -87.5, -8.4, 2.495, 20, unit.AngularMOA]
        ]

        for d in test_data:
            with self.subTest():
                self.validate_one(*d)

    def test_path_g7(self):
        bc = BallisticCoefficient(0.223, DragTableG7)
        projectile = ProjectileWithDimensions(bc, unit.Distance(0.308, unit.DistanceInch).validate(),
                                              unit.Distance(1.282, unit.DistanceInch).validate(),
                                              unit.Weight(168, unit.WeightGrain).validate())
        ammo = Ammunition(projectile, unit.Velocity(2750, unit.VelocityFPS).validate())
        zero = ZeroInfo(unit.Distance(100, unit.DistanceYard).validate())
        twist = TwistInfo(TwistRight, unit.Distance(11.24, unit.DistanceInch).validate())
        weapon = Weapon.create_with_twist(unit.Distance(2, unit.DistanceInch).validate(), zero, twist)
        atmosphere = Atmosphere()
        shot_info = ShotParameters(unit.Angular(4.221, unit.AngularMOA).validate(),
                                   unit.Distance(1000, unit.DistanceYard).validate(),
                                   unit.Distance(100, unit.DistanceYard).validate())
        wind = WindInfo.create_only_wind_info(unit.Velocity(5, unit.VelocityMPH).validate(),
                                              unit.Angular(-45, unit.AngularDegree).validate())

        calc = TrajectoryCalculator()
        data = calc.trajectory(ammo, weapon, atmosphere, shot_info, wind)

        self.assertEqualCustom(len(data), 11, 0.1, "Length")

        test_data = [
            [data[0], 0, 2750, 2.463, 2820.6, -2, 0, 0, 0, 0, 880, unit.AngularMil],
            [data[1], 100, 2544.3, 2.279, 2416, 0, 0, -0.35, -0.09, 0.113, 698, unit.AngularMil],
            [data[5], 500, 1810.7, 1.622, 1226, -56.3, -3.18, -9.96, -0.55, 0.673, 252, unit.AngularMil],
            [data[10], 1000, 1081.3, 0.968, 442, -401.6, -11.32, -50.98, -1.44, 1.748, 55, unit.AngularMil]
        ]

        for d in test_data:
            with self.subTest():
                self.validate_one(*d)


if __name__ == '__main__':
    unittest.main()



# import atmosphere
# import drag
# import projectile
# import shot_parameters
# import trajectory_calculator
# import trajectory_data
# import weapon
# import wind
# import bmath

__author__ = "o-murphy"
__copyright__ = ("",)

__credits__ = ["O-Murphy"]
__version__ = "1.0.0"

__all__ = ['atmosphere', 'bmath', 'drag', 'projectile', 'py_ballisticcalc_test', 'shot_parameters',
           'trajectory_calculator', 'weapon', 'wind']



from ..unit.convertor import Convertor

# the value indicating that distance value is expressed in some unit
DistanceInch = 10
DistanceFoot = 11
DistanceYard = 12
DistanceMile = 13
DistanceNauticalMile = 14
DistanceMillimeter = 15
DistanceCentimeter = 16
DistanceMeter = 17
DistanceKilometer = 18
DistanceLine = 19


class Distance(Convertor):
    """ Distance structure keeps the The distance value """

    __name__ = 'Distance'

    _units = {
        DistanceInch: {'name': 'in', 'accuracy': 1,
                       'to': lambda v: v,
                       'from': lambda v: v},
        DistanceFoot: {'name': 'ft', 'accuracy': 2,
                       'to': lambda v: v * 12,
                       'from': lambda v: v / 12},
        DistanceYard: {'name': 'yd', 'accuracy': 3,
                       'to': lambda v: v * 36,
                       'from': lambda v: v / 36},
        DistanceMile: {'name': 'mi', 'accuracy': 3,
                       'to': lambda v: v * 63360,
                       'from': lambda v: v / 63360},
        DistanceNauticalMile: {'name': 'nm', 'accuracy': 3,
                               'to': lambda v: v * 72913.3858,
                               'from': lambda v: v / 72913.3858},
        DistanceLine: {'name': 'ln', 'accuracy': 1,
                       'to': lambda v: v / 10,
                       'from': lambda v: v * 10},
        DistanceMillimeter: {'name': 'mm', 'accuracy': 0,
                             'to': lambda v: v / 25.4,
                             'from': lambda v: v * 25.4},
        DistanceCentimeter: {'name': 'cm', 'accuracy': 1,
                             'to': lambda v: v / 2.54,
                             'from': lambda v: v * 2.54},
        DistanceMeter: {'name': 'm', 'accuracy': 2,
                        'to': lambda v: v / 25.4 * 1000,
                        'from': lambda v: v * 25.4 / 1000},
        DistanceKilometer: {'name': 'km', 'accuracy': 3,
                            'to': lambda v: v / 25.4 * 1000000,
                            'from': lambda v: v * 25.4 / 1000000},
    }



from ..unit.convertor import Convertor

# the value indicating that pressure value is expressed in some unit
PressureMmHg = 40
PressureInHg = 41
PressureBar = 42
PressureHP = 43
PressurePSI = 44


class Pressure(Convertor):
    """ Pressure object keeps velocity or speed values """

    __name__ = 'Pressure'

    _units = {
        PressureMmHg: {'name': 'mmHg', 'accuracy': 0,
                       'to': lambda v: v,
                       'from': lambda v: v},
        PressureInHg: {'name': 'inHg', 'accuracy': 2,
                       'to': lambda v: v * 25.4,
                       'from': lambda v: v / 25.4},
        PressureBar: {'name': 'bar', 'accuracy': 2,
                      'to': lambda v: v * 750.061683,
                      'from': lambda v: v / 750.061683},
        PressureHP: {'name': 'hPa', 'accuracy': 4,
                     'to': lambda v: v * 750.061683 / 1000,
                     'from': lambda v: v / 750.061683 * 1000},
        PressurePSI: {'name': 'psi', 'accuracy': 4,
                      'to': lambda v: v * 51.714924102396,
                      'from': lambda v: v / 51.714924102396}
    }



import math
import unittest
from py_ballisticcalc.bmath import Vector


class TestVectorCreation(unittest.TestCase):

    def test_create(self):
        v = Vector(1, 2, 3)

        self.assertTrue(v.x == 1 and v.y == 2 and v.z == 3, "Create failed")

        c = v.__copy__()
        self.assertTrue(c.x == 1 and c.y == 2 and c.z == 3, "Copy failed")


class TestUnary(unittest.TestCase):

    def test_unary(self):
        v1 = Vector(1, 2, 3)

        self.assertFalse(math.fabs(v1.magnitude() - 3.74165738677) > 1e-7, "Magnitude failed")

        v2 = v1.negate()
        self.assertTrue(v2.x == -1 and v2.y == -2 and v2.z == -3, "Negate failed")

        v2 = v1.normalize()
        self.assertFalse(v2.x > 1 or v2.y > 1 or v2.z > 1, "Normalize failed")

        v1 = Vector(0, 0, 0)
        v2 = v1.normalize()
        self.assertFalse(v2.x != 0 or v2.y != 0 or v2.z != 0, "Normalize failed")


class TestBinary(unittest.TestCase):

    def test_binary(self):
        v1 = Vector(1, 2, 3)

        v2 = v1.add(v1.__copy__())
        self.assertFalse(v2.x != 2 or v2.y != 4 or v2.z != 6, "Add failed")

        v2 = v1.subtract(v2)
        self.assertFalse(v2.x != -1 or v2.y != -2 or v2.z != -3, "Subtract failed")

        self.assertFalse(v1.multiply_by_vector(v1.__copy__()) != (1 + 4 + 9), "MultiplyByVector failed")

        v2 = v1.multiply_by_const(3)
        self.assertFalse(v2.x != 3 or v2.y != 6 or v2.z != 9, "MultiplyByConst failed")



from py_ballisticcalc.projectile import *
from py_ballisticcalc.drag import *
from py_ballisticcalc.weapon import *
from py_ballisticcalc.trajectory_calculator import *
from py_ballisticcalc.atmosphere import *
from py_ballisticcalc.shot_parameters import *
from py_ballisticcalc.bmath import unit

bc = BallisticCoefficient(0.223, DragTableG7)
projectile = ProjectileWithDimensions(bc, unit.Distance(0.308, unit.DistanceInch).validate(),
                                      unit.Distance(1.282, unit.DistanceInch).validate(),
                                      unit.Weight(168, unit.WeightGrain).validate())
ammo = Ammunition(projectile, unit.Velocity(2750, unit.VelocityFPS).validate())
zero = ZeroInfo(unit.Distance(100, unit.DistanceMeter).validate())
twist = TwistInfo(TwistRight, unit.Distance(11.24, unit.DistanceInch).validate())
weapon = Weapon.create_with_twist(unit.Distance(2, unit.DistanceInch).validate(), zero, twist)
atmosphere = Atmosphere()
shot_info = ShotParameters(unit.Angular(4.221, unit.AngularMOA).validate(),
                           unit.Distance(1001, unit.DistanceMeter).validate(),
                           unit.Distance(100, unit.DistanceMeter).validate())
wind = WindInfo.create_only_wind_info(unit.Velocity(5, unit.VelocityMPH).validate(),
                                      unit.Angular(-45, unit.AngularDegree).validate())

calc = TrajectoryCalculator()
data = calc.trajectory(ammo, weapon, atmosphere, shot_info, wind)

for d in data:
    distance = d.travelled_distance
    meters = distance.convert(unit.DistanceMeter)
    velocity = d.velocity.convert(unit.VelocityMPS)
    mach = round(d.mach_velocity, 4)
    energy = d.energy
    time = round(d.time.total_seconds, 4)
    ogv = d.optimal_game_weight.get_in(unit.WeightPound)
    path = d.drop.convert(unit.DistanceCentimeter)
    hold = d.drop_adjustment.get_in(unit.AngularMOA) if distance.v > 1 else None
    windage = d.windage.convert(unit.DistanceCentimeter)
    wind_adjustment = d.windage_adjustment.get_in(unit.AngularMOA) if distance.v > 1 else None
    print(
        f'Distance: {meters}, Velocity: {velocity}, '
        f'Mach: {mach}, Energy: {energy}, '
        f'Time: {time}s, Path: {path}, Windage: {windage}'
    )



import math
from .bmath import unit


cIcaoStandardTemperatureR = 518.67
cIcaoFreezingPointTemperatureR = 459.67
cTemperatureGradient = -3.56616e-03
cIcaoStandardHumidity = 0.0
cPressureExponent = -5.255876
cSpeedOfSound = 49.0223
cA0 = 1.24871
cA1 = 0.0988438
cA2 = 0.00152907
cA3 = -3.07031e-06
cA4 = 4.21329e-07
cA5 = 3.342e-04
cStandardTemperature = 59.0
cStandardPressure = 29.92
cStandardDensity = 0.076474


class Atmosphere(object):
    """ Atmosphere describes the atmosphere conditions """

    _density: float = None
    _mach: [unit.Velocity, float] = None
    _mach1: float = None

    def __init__(self, altitude: unit.Distance = None, pressure: unit.Pressure = None,
                 temperature: unit.Temperature = None, humidity: float = 0.78):
        """
        Creates the atmosphere with the specified parameter
        :param altitude: unit.Distance instance
        :param pressure: unit.Pressure instance
        :param temperature: unit.Temperature instance
        :param humidity: 0 - 1 or 1 - 100 float
        """
        if humidity > 1:
            humidity = humidity / 100

        if humidity < 0 or humidity > 100:
            self.create_default()
        elif not altitude or not pressure or not temperature:
            self.create_default()
        else:
            self._altitude = altitude
            self._pressure = pressure
            self._temperature = temperature
            self._humidity = humidity

        self.calculate()

    def __str__(self) -> str:
        """
        :return: formatted Atmosphere data
        """
        return f'Altitude: {self._altitude}, Pressure: {self._pressure}, ' \
               f'Temperature: {self._temperature}, Humidity: {self.humidity_in_percent:.2f}'

    def create_default(self):
        self._altitude = unit.Distance(0.0, unit.DistanceFoot).validate()
        self._pressure = unit.Pressure(cStandardPressure, unit.PressureInHg).validate()
        self._temperature = unit.Temperature(cStandardTemperature, unit.TemperatureFahrenheit).validate()
        self._humidity = 0.78

    @property
    def altitude(self) -> unit.Distance:
        """
        :return: the ground level altitude over the sea level
        """
        return self._altitude

    @property
    def temperature(self) -> unit.Temperature:
        """
        :return: the temperature at the ground level
        """
        return self._temperature

    @property
    def pressure(self) -> unit.Pressure:
        """
        :return: the pressure at the ground level
        """
        return self._pressure

    @property
    def humidity(self) -> float:
        """
        :return: the relative humidity set in 0 to 1 coefficient
        multiply this value by 100 to get percents
        """
        return self._humidity

    @property
    def humidity_in_percent(self) -> float:
        """
        :return: relative humidity in percents (0..100)
        """
        return self._humidity * 100

    @property
    def density(self) -> float:
        return self._density

    @property
    def density_factor(self) -> float:
        return self._density / cStandardDensity

    @property
    def mach(self) -> [unit.Velocity, float]:
        """
        :return: the speed of sound at the atmosphere with such parameters
        """
        return self._mach

    def calculate0(self, t, p) -> tuple[float, float]:
        if t > 0:
            et0 = cA0 + t * (cA1 + t * (cA2 + t * (cA3 + t * cA4)))
            et = cA5 * self._humidity * et0
            hc = (p - 0.3783 * et) / cStandardPressure
        else:
            hc = 1.0

        density = cStandardDensity * (cIcaoStandardTemperatureR / (t + cIcaoFreezingPointTemperatureR)) * hc
        mach = math.sqrt(t + cIcaoFreezingPointTemperatureR) * cSpeedOfSound
        return density, mach

    def calculate(self) -> None:
        t = self._temperature.get_in(unit.TemperatureFahrenheit)
        p = self._pressure.get_in(unit.PressureInHg)
        density, mach = self.calculate0(t, p)
        self._density = density
        self._mach1 = mach
        self._mach = unit.Velocity(mach, unit.VelocityFPS).validate()

    def get_density_factor_and_mach_for_altitude(self, altitude: float) -> tuple[float, float]:
        org_altitude = self._altitude.get_in(unit.DistanceFoot)
        if math.fabs(org_altitude - altitude) < 30:
            density = self._density / cStandardDensity
            mach = self._mach1
            return density, mach

        t0 = self._temperature.get_in(unit.TemperatureFahrenheit)
        p = self._pressure.get_in(unit.PressureInHg)

        ta = cIcaoStandardTemperatureR + org_altitude * cTemperatureGradient - cIcaoFreezingPointTemperatureR
        tb = cIcaoStandardTemperatureR + altitude * cTemperatureGradient - cIcaoFreezingPointTemperatureR
        t = t0 + ta - tb
        p = p * math.pow(t0 / t, cPressureExponent)

        density, mach = self.calculate0(t, p)
        return density / cStandardDensity, mach

    @staticmethod
    def create_icao(altitude: unit.Distance):
        temperature = unit.Temperature(
            cIcaoStandardTemperatureR + altitude.get_in(
                unit.DistanceFoot
            ) * cTemperatureGradient - cIcaoFreezingPointTemperatureR, unit.TemperatureFahrenheit
        ).validate()

        pressure = unit.Pressure(
            cStandardPressure *
            math.pow(cIcaoStandardTemperatureR / (
                        temperature.get_in(unit.TemperatureFahrenheit) + cIcaoFreezingPointTemperatureR),
                     cPressureExponent),
            unit.PressureInHg
        ).validate()

        return Atmosphere(altitude, pressure, temperature, cIcaoStandardHumidity)


if __name__ == '__main__':

    atmo = Atmosphere(
        unit.Distance(500, unit.DistanceMeter),
        unit.Pressure(500, unit.PressureMmHg),
        unit.Temperature(26, unit.TemperatureCelsius),
        humidity=0.5  # 50%
    )

    # get speed of sound in mps at the atmosphere with such parameters
    speed_of_sound_in_mps = unit.Velocity(atmo.mach.get_in(unit.VelocityMPS), unit.VelocityMPS)



import math
import unittest
from py_ballisticcalc.bmath import unit


def test_back_n_forth(test, value, units):
    u = test.unit_class(value, units)
    test.assertEqual(u.error, None, f'Creation failed for {units}')

    v, err = u.value(units)
    test.assertTrue(
        err is None
        and math.fabs(v - value) < 1e-7
        and math.fabs(v - u.get_in(units) < 1e-7), f'Read back failed for {units}')


class TestAngular(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = unit.Angular
        self.unit_list = [
            unit.AngularDegree,
            unit.AngularMOA,
            unit.AngularMRad,
            unit.AngularMil,
            unit.AngularRadian,
            unit.AngularThousand
        ]

    def test_angular(self):
        for u in self.unit_list:
            with self.subTest(unit=unit):
                test_back_n_forth(self, 3, u)


class TestDistance(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = unit.Distance
        self.unit_list = [
            unit.DistanceCentimeter,
            unit.DistanceFoot,
            unit.DistanceInch,
            unit.DistanceKilometer,
            unit.DistanceLine,
            unit.DistanceMeter,
            unit.DistanceMillimeter,
            unit.DistanceMile,
            unit.DistanceNauticalMile,
            unit.DistanceYard
        ]

    def test_distance(self):
        for u in self.unit_list:
            with self.subTest(unit=unit):
                test_back_n_forth(self, 3, u)


class TestEnergy(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = unit.Energy
        self.unit_list = [
            unit.EnergyFootPound,
            unit.EnergyJoule
        ]

    def test_energy(self):
        for u in self.unit_list:
            with self.subTest(unit=unit):
                test_back_n_forth(self, 3, u)


class TestPressure(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = unit.Pressure
        self.unit_list = [
            unit.PressureBar,
            unit.PressureHP,
            unit.PressureMmHg,
            unit.PressureInHg
        ]

    def test_pressure(self):
        for u in self.unit_list:
            with self.subTest(unit=unit):
                test_back_n_forth(self, 3, u)


class TestTemperature(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = unit.Temperature
        self.unit_list = [
            unit.TemperatureFahrenheit,
            unit.TemperatureKelvin,
            unit.TemperatureCelsius,
            unit.TemperatureRankin
        ]

    def test_temperature(self):
        for u in self.unit_list:
            with self.subTest(unit=unit):
                test_back_n_forth(self, 3, u)


class TestVelocity(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = unit.Velocity
        self.unit_list = [
            unit.VelocityFPS,
            unit.VelocityKMH,
            unit.VelocityKT,
            unit.VelocityMPH,
            unit.VelocityMPS
        ]

    def test_velocity(self):
        for u in self.unit_list:
            with self.subTest(unit=unit):
                test_back_n_forth(self, 3, u)


class TestWeight(unittest.TestCase):
    def setUp(self) -> None:
        self.unit_class = unit.Weight
        self.unit_list = [
            unit.WeightGrain,
            unit.WeightGram,
            unit.WeightKilogram,
            unit.WeightNewton,
            unit.WeightOunce,
            unit.WeightPound
        ]

    def test_weight(self):
        for u in self.unit_list:
            with self.subTest(unit=unit):
                test_back_n_forth(self, 3, u)


if __name__ == '__main__':
    unittest.main()



from ..unit.convertor import Convertor

# the value indicating that weight value is expressed in some unit
WeightGrain = 70
WeightOunce = 71
WeightGram = 72
WeightPound = 73
WeightKilogram = 74
WeightNewton = 75

#
# class WeightConvertor(UnitsConvertor):
#     unit_type = 'weight'
#
#     _units = {
#         WeightGrain: {'name': 'grn', 'accuracy': 0,
#                       'to': lambda v: v,
#                       'from': lambda v: v},
#         WeightGram: {'name': 'grn', 'accuracy': 1,
#                      'to': lambda v: v * 15.4323584,
#                      'from': lambda v: v / 15.4323584},
#         WeightKilogram: {'name': 'grn', 'accuracy': 3,
#                          'to': lambda v: v * 15432.3584,
#                          'from': lambda v: v / 15432.3584},
#         WeightNewton: {'name': 'grn', 'accuracy': 3,
#                        'to': lambda v: v * 151339.73750336,
#                        'from': lambda v: v / 151339.73750336},
#         WeightPound: {'name': 'grn', 'accuracy': 3,
#                       'to': lambda v: v / 0.000142857143,
#                       'from': lambda v: v * 0.000142857143},
#         WeightOunce: {'name': 'grn', 'accuracy': 1,
#                       'to': lambda v: v * 437.5,
#                       'from': lambda v: v / 437.5},
#     }


class Weight(Convertor):
    """ Weight object keeps data about weight """
    # convertor = WeightConvertor

    __name__ = 'Weight'

    _units = {
        WeightGrain: {'name': 'grn', 'accuracy': 0,
                      'to': lambda v: v,
                      'from': lambda v: v},
        WeightGram: {'name': 'grn', 'accuracy': 1,
                     'to': lambda v: v * 15.4323584,
                     'from': lambda v: v / 15.4323584},
        WeightKilogram: {'name': 'grn', 'accuracy': 3,
                         'to': lambda v: v * 15432.3584,
                         'from': lambda v: v / 15432.3584},
        WeightNewton: {'name': 'grn', 'accuracy': 3,
                       'to': lambda v: v * 151339.73750336,
                       'from': lambda v: v / 151339.73750336},
        WeightPound: {'name': 'grn', 'accuracy': 3,
                      'to': lambda v: v / 0.000142857143,
                      'from': lambda v: v * 0.000142857143},
        WeightOunce: {'name': 'grn', 'accuracy': 1,
                      'to': lambda v: v * 437.5,
                      'from': lambda v: v / 437.5},
    }

    def __init__(self, value: float, units: int):
        super(Weight, self).__init__(value, units)



import math


class Vector(object):
    """ Vector object keeps data about a 3D vector """

    def __init__(self, x: float, y: float, z: float):
        """
        Create create a vector from its coordinates
        :param x: X-coordinate
        :param y: Y-coordinate
        :param z: Z-coordinate
        """
        self.x = x
        self.y = y
        self.z = z

    def __str__(self) -> str:
        """
        Converts a vector into a string
        :return: formatted string [X, Y, Z]
        """
        return f'[X={self.x}, Y={self.y}, Z={self.z}]'

    def __copy__(self) -> 'Vector':
        """
        Creates a copy of the vector
        :return: Vector
        """
        return Vector(self.x, self.y, self.z)

    def multiply_by_vector(self, v2: 'Vector') -> float:
        """
        Returns a product of two vectors
        The product of two vectors is a sum of products of each coordinate
        :param v2: Vector(x2, y2, z2)
        :return: float
        """
        return self.x * v2.x + self.y * v2.y + self.z * v2.z

    def magnitude(self) -> float:
        """
        Returns a magnitude of the vector
        The magnitude of the vector is the length of a line that starts in point (0,0,0)
        and ends in the point set by the vector coordinates
        :return: magnitude of the vector
        """
        return math.sqrt(math.pow(self.x, 2) + math.pow(self.y, 2) + math.pow(self.z, 2))

    def multiply_by_const(self, a: float) -> 'Vector':
        """
        Multiplies the vector by the constant
        :param a: float multiplier
        :return: Vector
        """
        return Vector(self.x * a, self.y * a, self.z * a)

    def add(self, v2: 'Vector') -> 'Vector':
        """
        Adds two vectors
        :param v2: Vector(x2, y2, z2)
        :return: sum of two vectors
        """
        return Vector(self.x + v2.x, self.y + v2.y, self.z + v2.z)

    def subtract(self, v2: 'Vector') -> 'Vector':
        """
        Subtracts one vector from another
        :param v2: Vector(x2, y2, z2)
        :return: Vector
        """
        return Vector(self.x - v2.x, self.y - v2.y, self.z - v2.z)

    def negate(self) -> 'Vector':
        """
        Returns a vector which is symmetrical to this vector vs (0,0,0) point
        :return: Vector
        """
        return Vector(-self.x, -self.y, -self.z)

    def normalize(self) -> 'Vector':
        """
        Returns a vector of magnitude one which is collinear to this vector
        :return: Vector
        """
        magnitude = self.magnitude()
        if math.fabs(magnitude) < 1e-10:
            return self.__copy__()
        return self.multiply_by_const(1.0 / magnitude)



from .bmath import unit
from .atmosphere import Atmosphere
from .projectile import Ammunition


class ZeroInfo(object):
    """ ZeroInfo object keeps the information about zeroing of the weapon """

    def __init__(self, distance: unit.Distance,
                 has_ammunition: bool = False,
                 has_atmosphere: bool = False,
                 ammunition: Ammunition = None, atmosphere: Atmosphere = None):
        self._has_ammunition = has_ammunition
        self._has_atmosphere = has_atmosphere
        self._zero_distance = distance
        self._ammunition = ammunition
        self._zero_atmosphere = atmosphere

    @property
    def has_ammunition(self) -> bool:
        """
        :return: flag indicating whether other ammo is used to zero
        """
        return self._has_ammunition

    @property
    def ammunition(self) -> Ammunition:
        """
        :return: ammo used to zero
        """
        return self._ammunition

    @property
    def has_atmosphere(self) -> bool:
        """
        :return: flag indicating whether weapon is zeroed under different conditions
        """
        return self._has_atmosphere

    @property
    def atmosphere(self) -> Atmosphere:
        """
        :return: conditions at the time of zeroing
        """
        return self._zero_atmosphere

    @property
    def zero_distance(self) -> unit.Distance:
        """
        :return: the distance at which the weapon was zeroed
        """
        return self._zero_distance

    @staticmethod
    def create_info(distance: unit.Distance) -> 'ZeroInfo':
        return ZeroInfo(has_ammunition=False,
                        has_atmosphere=False,
                        distance=distance)

    @staticmethod
    def create_with_atmosphere(distance: unit.Distance, atmosphere: Atmosphere) -> 'ZeroInfo':
        """
        Creates zero information using distance and conditions
        :param distance: unit.Distance instance
        :param atmosphere: Atmosphere instance
        :return: ZeroInfo instance
        """
        return ZeroInfo(has_ammunition=False,
                        has_atmosphere=True,
                        distance=distance,
                        atmosphere=atmosphere)

    @staticmethod
    def create_with_another_ammo(distance: unit.Distance, ammo: Ammunition) -> 'ZeroInfo':
        """
        Creates zero information using distance and other ammunition
        :param distance: unit.Distance instance
        :param ammo: Ammunition instance
        :return: ZeroInfo instance
        """
        return ZeroInfo(has_ammunition=True,
                        has_atmosphere=False,
                        distance=distance,
                        ammunition=ammo)

    @staticmethod
    def create_with_another_ammo_and_atmosphere(distance: unit.Distance, ammo: Ammunition,
                                                atmosphere: Atmosphere) -> 'ZeroInfo':
        """
        Creates zero information using distance, other conditions and other ammunition
        :param distance: unit.Distance instance
        :param ammo: Ammunition instance
        :param atmosphere: Atmosphere instance
        :return: ZeroInfo instance
        """
        return ZeroInfo(
            has_ammunition=True,
            has_atmosphere=False,
            distance=distance,
            ammunition=ammo,
            atmosphere=atmosphere
        )


TwistRight = 1
TwistLeft = 2


class TwistInfo(object):
    def __init__(self, direction: int, twist: unit.Distance):
        """
        Creates twist information
        :param direction: Direction must be either Twist_Right or Twist_Left constant
        :param twist: unit.Distance instance
        """
        self._twist_direction = direction
        self._rifling_twist = twist

    @property
    def direction(self) -> int:
        """
        :return: the twist direction (see TwistRight and TwistLeft)
        """
        return self._twist_direction

    @property
    def twist(self) -> unit.Distance:
        """
        :return: the twist step (the distance inside the barrel at which the projectile makes one turn)
        """
        return self._rifling_twist


class Weapon(object):
    """ Weapon object contains the weapon description """

    def __init__(self, sight_height: unit.Distance, zero_info: ZeroInfo,
                 has_twist_info: bool = False, twist: TwistInfo = None, click_value: unit.Angular = None):
        """
        Ceates the weapon definition with no twist info
        :param sight_height: unit.Distance instance
        :param zero_info: ZeroInfo instance
        :param has_twist_info: bool (is twist)
        :param twist: TwistInfo instance
        :param click_value: unit.Angular instance
        """
        self._sight_height = sight_height
        self._zero_info = zero_info
        self._has_twist_info = has_twist_info
        self._twist = twist
        self._click_value = click_value

    @property
    def sight_height(self) -> unit.Distance:
        """
        :return: the height of the sight center line over the barrel center line
        """
        return self._sight_height

    @property
    def zero(self) -> ZeroInfo:
        """
        :return: the zeroing information
        """
        return self._zero_info

    @property
    def has_twist(self) -> bool:
        """
        :return: the flag indicating whether the rifling twist information is set
        """
        return self._has_twist_info

    @property
    def twist(self) -> TwistInfo:
        """
        :return: the rifling twist information
        """
        return self._twist

    @property
    def click_value(self) -> unit.Angular:
        """
        :return: the value of one click of the scope
        """
        return self._click_value

    @click_value.setter
    def click_value(self, click: unit.Angular):
        """
        Sets the value of one click of the scope
        :param click: unit.Angular instance
        :return: None
        """
        self._click_value = click

    @staticmethod
    def create_with_twist(sight_height: unit.Distance, zero_info: ZeroInfo, twist: TwistInfo) -> 'Weapon':
        """
        Creates weapon description with twist info
        :param sight_height: unit.Distance instance
        :param zero_info: ZeroInfo instance
        :param twist: TwistInfo instance
        :return: Weapon instance
        """
        return Weapon(
            sight_height=sight_height,
            zero_info=zero_info,
            has_twist_info=True,
            twist=twist
        )



from .bmath import unit
from .drag import BallisticCoefficient


class Projectile(object):
    _bullet_diameter: unit.Distance = None
    _bullet_length: unit.Distance = None

    """
    keeps description of a projectile
    """
    def __init__(self, ballistic_coefficient: BallisticCoefficient, weight: unit.Weight):
        """
        projectile description with dimensions
        :param ballistic_coefficient: BallisticCoefficient instance
        :param weight: unit.Weight instance
        """
        self._ballistic_coefficient = ballistic_coefficient
        self._has_dimensions = False
        self._weight = weight

    @property
    def ballistic_coefficient(self) -> BallisticCoefficient:
        """
        :return: ballistic coefficient of the projectile
        """
        return self._ballistic_coefficient

    @property
    def bullet_weight(self) -> unit.Weight:
        """
        :return: weight of the projectile
        """
        return self._weight

    @property
    def bullet_diameter(self) -> unit.Distance:
        """
        :return: the diameter (caliber) of the projectile
        """
        return self._bullet_diameter

    @property
    def bullet_length(self) -> unit.Distance:
        """
        :return: the length of the bullet
        """
        return self._bullet_length

    @property
    def has_dimensions(self) -> bool:
        """
        :return: flag indicating whether the projectile
        """
        return self._has_dimensions


class ProjectileWithDimensions(Projectile):
    """
    ProjectileWithDimensions creates the description of a projectile with dimensions (diameter and length)
    Dimensions are only required if you want to take into account projectile spin drift.
    TwistInfo must be also set in this case.
    """

    def __init__(self, ballistic_coefficient: BallisticCoefficient,
                 bullet_diameter: unit.Distance,
                 bullet_length: unit.Distance,
                 weight: unit.Weight):
        """
        :param ballistic_coefficient: BallisticCoefficient instance
        :param bullet_diameter: unit.Distance instance
        :param bullet_length: unit.Distance instance
        :param weight: unit.Weight instance
        """

        super(ProjectileWithDimensions, self).__init__(ballistic_coefficient, weight)
        self._has_dimensions = True
        self._bullet_diameter = bullet_diameter
        self._bullet_length = bullet_length


class Ammunition(object):
    """ Ammunition object keeps the des of ammunition (e.g. projectile loaded into a case shell) """

    def __init__(self, bullet: Projectile, muzzle_velocity: unit.Velocity):
        """
        creates the description of the ammunition
        :param bullet: Projectile instance
        :param muzzle_velocity: unit.Velocity instance
        """
        self._projectile = bullet
        self._muzzle_velocity = muzzle_velocity

    @property
    def bullet(self) -> Projectile:
        """
        :return: the description of the projectile
        """
        return self._projectile

    @property
    def muzzle_velocity(self) -> unit.Velocity:
        """
        :return: the velocity of the projectile at the muzzle
        """
        return self._muzzle_velocity



from ..unit.convertor import Convertor

# the value indicating that velocity value is expressed in some unit
VelocityMPS = 60
VelocityKMH = 61
VelocityFPS = 62
VelocityMPH = 63
VelocityKT = 64


class Velocity(Convertor):
    """ Velocity object keeps velocity or speed values """

    __name__ = 'Velocity'

    _units = {
        VelocityMPS: {'name': 'm/s', 'accuracy': 0,
                      'to': lambda v: v,
                      'from': lambda v: v},
        VelocityKMH: {'name': 'km/h', 'accuracy': 1,
                      'to': lambda v: v / 3.6,
                      'from': lambda v: v * 3.6},
        VelocityFPS: {'name': 'ft/s', 'accuracy': 1,
                      'to': lambda v: v / 3.2808399,
                      'from': lambda v: v * 3.2808399},
        VelocityMPH: {'name': 'mph', 'accuracy': 1,
                      'to': lambda v: v / 2.23693629,
                      'from': lambda v: v * 2.23693629},
        VelocityKT: {'name': 'kt', 'accuracy': 1,
                     'to': lambda v: v / 1.94384449,
                     'from': lambda v: v * 1.94384449},
    }



from .bmath import unit


class ShotParameters(object):
    def __init__(self, sight_angle: unit.Angular, maximum_distance: unit.Distance, step: unit.Distance):
        """
        Creates parameters of the shot
        :param sight_angle: the angle between scope center line and the barrel center line
        :param maximum_distance:
        :param step:
        """
        self._sight_angle = sight_angle
        self._shot_angle = unit.Angular(0, unit.AngularRadian).validate()
        self._cant_angle = unit.Angular(0, unit.AngularRadian).validate()
        self._maximum_distance: unit.Distance = maximum_distance
        self._step: unit.Distance = step

    @property
    def sight_angle(self) -> unit.Angular:
        """
        :return: the angle of the sight
        """
        return self._sight_angle

    @property
    def shot_angle(self) -> unit.Angular:
        """
        :return: the angle of the shot
        """
        return self._shot_angle

    @property
    def cant_angle(self):
        """
        :return: the cant angle (the angle between centers of scope and the barrel projection and zenith line)
        """
        return self._cant_angle

    @property
    def maximum_distance(self) -> unit.Distance:
        """
        :return: the maximum distance to be calculated
        """
        return self._maximum_distance

    @property
    def step(self) -> unit.Distance:
        """
        :return: the step between calculation results
        """
        return self._step


class ShotParameterUnlevel(ShotParameters):
    def __init__(self, sight_angle: unit.Angular, maximum_distance: unit.Distance,
                 step: unit.Distance, shot_angle: unit.Angular, cant_angle: unit.Angular):
        super(ShotParameterUnlevel, self).__init__(sight_angle, maximum_distance, step)
        self._sight_angle = sight_angle
        self._shot_angle = shot_angle
        self._cant_angle = cant_angle
        self._maximum_distance: unit.Distance = maximum_distance
        self._step: unit.Distance = step



from .bmath import unit


class WindInfo(object):
    """ WindInfo structure keeps information about wind """

    def __init__(self, until_distance: unit.Distance = None,
                 velocity: unit.Velocity = None, direction: unit.Angular = None):
        self._until_distance = until_distance
        self._velocity = velocity
        self._direction = direction

    def __str__(self):
        """
        :return: formatted Wind data
        """
        return f'Until distance: {self._until_distance}, Velocity: {self._velocity}, Direction: {self._direction}'

    @property
    def until_distance(self) -> unit.Distance:
        """
        :return: the distance from the shooter until which the wind blows
        """
        return self._until_distance

    @property
    def velocity(self) -> unit.Velocity:
        """
        :return: the wind velocity
        """
        return self._velocity

    @property
    def direction(self) -> unit.Angular:
        """
        0 degrees means wind blowing into the face
        90 degrees means wind blowing from the left
        -90 or 270 degrees means wind blowing from the right
        180 degrees means wind blowing from the back
        :return: the wind direction
        """
        return self._direction

    @staticmethod
    def create_no_wind() -> list['WindInfo']:
        """
        Creates wind description with no wind
        :return: list[WindInfo]
        """
        return [WindInfo()]

    @staticmethod
    def create_only_wind_info(wind_velocity: unit.Velocity, direction: unit.Angular) -> list['WindInfo']:
        """
        Creates the wind information for the constant wind for the whole distance of the shot
        :param wind_velocity: unit.Velocity instance
        :param direction: unit.Angular instance
        :return: list[WindInfo]
        """
        until_distance = unit.Distance(9999, unit.DistanceKilometer)
        return [WindInfo(until_distance, wind_velocity, direction)]

    @staticmethod
    def add_wind_info(until_distance: unit.Distance,
                      velocity: unit.Velocity, direction: unit.Angular) -> 'WindInfo':
        """
        Creates description of one wind
        :param until_distance: unit.Distance instance
        :param velocity: unit.Velocity instance
        :param direction: unit.Angular instance
        :return: WindInfo instance
        """
        return WindInfo(until_distance, velocity, direction)

    @staticmethod
    def create_wind_info(*winds: 'WindInfo') -> list['WindInfo']:
        """
        Creates a wind descriptor from multiple winds
        winds must be ordered from the closest to the muzzle point to the farthest to the muzzle point
        :param winds: *parsed list or ruple of WindInfo instances
        :return: list[WindInfo]
        """
        return list(winds)


if __name__ == '__main__':
    wind = WindInfo(
        unit.Distance(500, unit.DistanceMeter),
        unit.Velocity(15, unit.VelocityKMH),
        unit.Angular(10, unit.AngularRadian)
    )



from .bmath import unit
import math


class Timespan(object):
    """ Timespan object keeps the amount of time spent """

    def __init__(self, time: float):
        self._time = time

    @property
    def total_seconds(self) -> float:
        """
        :return: the total number of seconds
        """
        return self._time

    @property
    def seconds(self) -> float:
        """
        :return: the whole number of the seconds
        """
        return math.fmod(math.floor(self._time), 60)

    @property
    def minutes(self) -> float:
        """
        :return: the whole number of minutes
        """
        return math.fmod(math.floor(self._time / 60), 60)


class TrajectoryData(object):
    """ TrajectoryData object keeps information about one point of the trajectory """

    def __init__(self, time: Timespan,
                 travel_distance: unit.Distance,
                 velocity: unit.Velocity,
                 mach: float,
                 drop: unit.Distance,
                 drop_adjustment: unit.Angular,
                 windage: unit.Distance,
                 windage_adjustment: unit.Angular,
                 energy: unit.Energy,
                 optimal_game_weight: unit.Weight):
        self._time = time
        self._travel_distance = travel_distance
        self._velocity = velocity
        self._mach = mach
        self._drop = drop
        self._drop_adjustment = drop_adjustment
        self._windage = windage
        self._windage_adjustment = windage_adjustment
        self._energy = energy
        self._optimal_game_weight = optimal_game_weight

    @property
    def time(self) -> Timespan:
        """
        :return: the amount of time spent since the shot moment
        """
        return self._time

    @property
    def travelled_distance(self) -> unit.Distance:
        """
        :return: the distance measured between the muzzle and the projection of the current bullet position to
        the line between the muzzle and the target
        """
        return self._travel_distance

    @property
    def velocity(self) -> unit.Velocity:
        """
        :return: the current projectile velocity
        """
        return self._velocity

    @property
    def mach_velocity(self) -> float:
        """
        :return: the proportion between the current projectile velocity and the speed of the sound
        """
        return self._mach

    @property
    def drop(self) -> unit.Distance:
        """
        :return: the shorted distance between the projectile and the shot line
        The positive value means the the projectile is above this line and the negative value means that the projectile
        is below this line
        """
        return self._drop

    @property
    def drop_adjustment(self) -> unit.Angular:
        """
        :return: the angle between the shot line and the line from the muzzle to the current projectile position
        in the plane perpendicular to the ground
        """
        return self._drop_adjustment

    @property
    def windage(self) -> unit.Distance:
        """
        :return: the distance to which the projectile is displaced by wind
        """
        return self._windage

    @property
    def windage_adjustment(self) -> unit.Angular:
        """
        :return: the angle between the shot line and the line from the muzzle to the current projectile position
        in the place parallel to the ground
        """
        return self._windage_adjustment

    @property
    def energy(self) -> unit.Energy:
        """
        :return: the kinetic energy of the projectile
        """
        return self._energy

    @property
    def optimal_game_weight(self) -> unit.Weight:
        """
        :return: the weight of game to which a kill shot is
        probable with the kinetic energy that the projectile currently  have
        """
        return self._optimal_game_weight



import math
# from ..unit.convertor import Convertor
from py_ballisticcalc.bmath.unit.convertor import Convertor

# the value indicating that angular value is expressed in some unit
AngularRadian = 0
AngularDegree = 1
AngularMOA = 2
AngularMil = 3
AngularMRad = 4
AngularThousand = 5
AngularInchesPer100Yd = 6
AngularCmPer100M = 7


class Angular(Convertor):
    """ Angular object keeps information about angular units """

    __name__ = 'Angular'

    _units = {
        AngularRadian: {'name': 'rad', 'accuracy': 6,
                        'to': lambda v: v,
                        'from': lambda v: v},
        AngularDegree: {'name': '°', 'accuracy': 4,
                        'to': lambda v: v / 180 * math.pi,
                        'from': lambda v: v * 180 / math.pi},
        AngularMOA: {'name': 'moa', 'accuracy': 2,
                     'to': lambda v: v / 180 * math.pi / 60,
                     'from': lambda v: v * 180 / math.pi * 60},
        AngularMil: {'name': 'mil', 'accuracy': 2,
                     'to': lambda v: v / 3200 * math.pi,
                     'from': lambda v: v * 3200 / math.pi},
        AngularMRad: {'name': 'mrad', 'accuracy': 2,
                      'to': lambda v: v / 1000,
                      'from': lambda v: v * 1000},
        AngularThousand: {'name': 'ths', 'accuracy': 2,
                          'to': lambda v: v / 3000 * math.pi,
                          'from': lambda v: v * 3000 / math.pi},
        AngularInchesPer100Yd: {'name': 'in/100yd', 'accuracy': 2,
                                'to': lambda v: math.atan(v / 3600),
                                'from': lambda v: math.tan(v) * 3600},
        AngularCmPer100M: {'name': 'cm/100m', 'accuracy': 2,
                           'to': lambda v: math.atan(v / 10000),
                           'from': lambda v: math.tan(v) * 10000},
    }



#!/usr/bin/env python
import sys
import py_ballisticcalc

from setuptools import setup, find_packages

setup(
    name='py_ballisticcalc',
    version=py_ballisticcalc.__version__,
    packages=find_packages(),
    url='https://github.com/o-murphy/py_ballisticcalc',
    download_url='http://pypi.python.org/pypi/py_ballisticcalc/',
    license='LGPL-3.0',
    author=py_ballisticcalc.__author__,
    author_email='thehelixpg@gmail.com',
    description='LGPL library for small arms ballistic calculations (Python 3)',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    zip_safe=True,
    py_modules=['py_ballisticcalc'],
    requirements=[]
)



from ..unit.convertor import Convertor

# the value indicating that energy value is expressed in some unit
EnergyFootPound = 30
EnergyJoule = 31


class Energy(Convertor):
    """ Energy object keeps information about kinetic energy """

    __name__ = 'Energy'

    _units = {
        EnergyFootPound: {'name': 'ft*lb', 'accuracy': 0,
                          'to': lambda v: v,
                          'from': lambda v: v},
        EnergyJoule: {'name': 'J', 'accuracy': 0,
                      'to': lambda v: v * 0.737562149277,
                      'from': lambda v: v / 0.737562149277}
    }



class Convertor(object):
    """ Basic units class, store units and value """

    __name__ = 'Convertor'

    _units = {}

    def __init__(self, value: float = None, units: int = None):
        """
        Creates a units value, strongly recommended validate it with validate() method
        usage example: Convertor(value, units).validate()
        :param value: units value
        :param units: UnitsConvertor constsS
        """
        v, err = self.to_default(value, units)
        if err:
            self._value = None
            self._default_units = units
        else:
            self._value = v
            self._default_units = units
        self.error = err

    def __str__(self) -> str:
        """
        :return: formatted value in default units
        """
        v, err = self.from_default(self._value, self._default_units)
        if err:
            return f'{self.__name__}: unit {self._default_units} is not supported'
        name, accuracy, f, t = self._units[self._default_units].values()
        return f'{round(v, accuracy)} {name}'

    def to_default(self, value: float, units: int) -> [float, Exception]:
        """
        :param value: value in specified units
        :param units: specified units
        :return: value in default units
        """
        try:
            return self._units[units]['to'](value), None
        except KeyError as error:
            return 0, f'KeyError: {self.__name__}: unit {error} is not supported'

    def from_default(self, value: float, units: int) -> [float, Exception]:
        """
        :param value: value in default units
        :param units: specified units
        :return: value in specified units
        """
        try:
            return self._units[units]['from'](value), None
        except KeyError as error:
            return 0, f'KeyError: {self.__name__}: unit {error} is not supported'

    def validate(self) -> 'Convertor':
        """
        usage example: Convertor(value, units).validate()
        required to use instead of Convertor(value, units) because it raise exception if something went wrong
        :return: the temperature value but panics instead of return error
        """
        err = self.error

        if err:
            err_msg = f'{self.__name__}: unit {self._default_units} is not supported'
            raise ValueError(err_msg)
        return self

    def value(self, units: int) -> [float, Exception]:
        """
        :param units: Units consts
        :return: Value of the unit in the specified units
        """
        return self.from_default(self._value, units)

    def convert(self, units: int) -> 'Convertor':
        """
        Returns the value into the specified units
        :param units: Units consts
        :return: Units object in the specified units
        """
        return self.__class__(self.get_in(units), units)

    def get_in(self, units: int) -> [float, Exception]:
        """
        Converts the value in the specified units.
        Returns 0 if unit conversion is not possible.
        :param units: Units consts
        :return: float
        """
        v, err = self.from_default(self._value, units)
        if err:
            return 0
        return v

    @property
    def v(self):
        """
        Almost useless but can be used for debug
        :return: current value in default units
        """
        return self._value

    @property
    def units(self):
        """
        :return: default units
        """
        return self._default_units
